@@ -58,7 +58,6 @@
 #include "textfile.h"
 
 #include "error.h"
-#include "storage.h"
 #include "sdpd.h"
 #include "glib-helper.h"
 #include "btio.h"
@@ -82,17 +81,19 @@ typedef enum {
 } proxy_type_t;
 
 struct serial_adapter {
-	bdaddr_t	src;		/* Adapter address */
-	char		*path;		/* D-Bus path */
-	DBusConnection	*conn;		/* Adapter connection */
-	GSList		*proxies;	/* Proxies list */
+	struct btd_adapter	*btd_adapter;	/* Adapter pointer */
+	DBusConnection		*conn;		/* Adapter connection */
+	GSList			*proxies;	/* Proxies list */
 };
 
 struct serial_proxy {
+	bdaddr_t	src;		/* Local address */
 	bdaddr_t	dst;		/* Remote address */
 	char		*path;		/* Proxy path */
 	char		*uuid128;	/* UUID 128 */
 	char		*address;	/* TTY or Unix socket name */
+	char		*owner;		/* Application bus name */
+	guint		watch;		/* Application watch */
 	short int	port;		/* TCP port */
 	proxy_type_t	type;		/* TTY or Unix socket */
 	struct termios  sys_ti;		/* Default TTY setting */
@@ -131,6 +132,7 @@ static void disable_proxy(struct serial_proxy *prx)
 
 static void proxy_free(struct serial_proxy *prx)
 {
+	g_free(prx->owner);
 	g_free(prx->path);
 	g_free(prx->address);
 	g_free(prx->uuid128);
@@ -189,6 +191,7 @@ static sdp_record_t *proxy_record_new(const char *uuid128, uint8_t channel)
 	sdp_list_free(root, NULL);
 
 	bt_string2uuid(&uuid, uuid128);
+	sdp_uuid128_to_uuid(&uuid);
 	svclass_id = sdp_list_append(NULL, &uuid);
 	sdp_set_service_classes(record, svclass_id);
 	sdp_list_free(svclass_id, NULL);
@@ -214,8 +217,7 @@ static sdp_record_t *proxy_record_new(const char *uuid128, uint8_t channel)
 
 	add_lang_attr(record);
 
-	sdp_set_info_attr(record, "Port Proxy Entity",
-				NULL, "Port Proxy Entity");
+	sdp_set_info_attr(record, "Serial Proxy", NULL, "Serial Proxy");
 
 	sdp_data_free(ch);
 	sdp_list_free(proto[0], NULL);
@@ -307,10 +309,11 @@ static inline int unix_socket_connect(const char *address)
 		 * Abstract namespace: first byte NULL, x00
 		 * must be removed from the original address.
 		 */
-		strcpy(addr.sun_path + 1, address + 3);
+		strncpy(addr.sun_path + 1, address + 3,
+						sizeof(addr.sun_path) - 2);
 	} else {
 		/* Filesystem address */
-		strcpy(addr.sun_path, address);
+		strncpy(addr.sun_path, address, sizeof(addr.sun_path) - 1);
 	}
 
 	/* Unix socket */
@@ -494,7 +497,7 @@ static void confirm_event_cb(GIOChannel *chan, gpointer user_data)
 
 	prx->rfcomm = g_io_channel_ref(chan);
 
-	perr = btd_request_authorization(&prx->adapter->src, &prx->dst,
+	perr = btd_request_authorization(&prx->src, &prx->dst,
 					prx->uuid128, auth_cb, prx);
 	if (perr < 0) {
 		error("Refusing connect from %s: %s (%d)", address,
@@ -510,30 +513,27 @@ drop:
 	g_io_channel_shutdown(chan, TRUE, NULL);
 }
 
-static DBusMessage *proxy_enable(DBusConnection *conn,
-				DBusMessage *msg, void *data)
+static int enable_proxy(struct serial_proxy *prx)
 {
-	struct serial_proxy *prx = data;
-	struct serial_adapter *adapter = prx->adapter;
 	sdp_record_t *record;
-	GError *err = NULL;
-	DBusMessage *reply;
+	GError *gerr = NULL;
+	int err;
 
 	if (prx->io)
-		return failed(msg, "Already enabled");
+		return -EALREADY;
 
 	/* Listen */
 	prx->io = bt_io_listen(BT_IO_RFCOMM, NULL, confirm_event_cb, prx,
-				NULL, &err,
-				BT_IO_OPT_SOURCE_BDADDR, &adapter->src,
+				NULL, &gerr,
+				BT_IO_OPT_SOURCE_BDADDR, &prx->src,
 				BT_IO_OPT_INVALID);
 	if (!prx->io)
 		goto failed;
 
-	bt_io_get(prx->io, BT_IO_RFCOMM, &err,
+	bt_io_get(prx->io, BT_IO_RFCOMM, &gerr,
 			BT_IO_OPT_CHANNEL, &prx->channel,
 			BT_IO_OPT_INVALID);
-	if (err) {
+	if (gerr) {
 		g_io_channel_unref(prx->io);
 		prx->io = NULL;
 		goto failed;
@@ -546,24 +546,42 @@ static DBusMessage *proxy_enable(DBusConnection *conn,
 	record = proxy_record_new(prx->uuid128, prx->channel);
 	if (!record) {
 		g_io_channel_unref(prx->io);
-		return failed(msg, "Unable to allocate new service record");
+		return -ENOMEM;
 	}
 
-	if (add_record_to_server(&adapter->src, record) < 0) {
+	err = add_record_to_server(&prx->src, record);
+	if (err < 0) {
 		sdp_record_free(record);
 		g_io_channel_unref(prx->io);
-		return failed(msg, "Service registration failed");
+		return err;
 	}
 
 	prx->record_id = record->handle;
 
-	return dbus_message_new_method_return(msg);
+	return 0;
 
 failed:
-	error("%s", err->message);
-	reply = failed(msg, err->message);
-	g_error_free(err);
-	return reply;
+	error("%s", gerr->message);
+	g_error_free(gerr);
+	return -EIO;
+
+}
+static DBusMessage *proxy_enable(DBusConnection *conn,
+				DBusMessage *msg, void *data)
+{
+	struct serial_proxy *prx = data;
+	int err;
+
+	err = enable_proxy(prx);
+	if (err == -EALREADY)
+		return failed(msg, "Already enabled");
+	else if (err == -ENOMEM)
+		return failed(msg, "Unable to allocate new service record");
+	else if (err < 0)
+		return g_dbus_create_error(msg, ERROR_INTERFACE "Failed",
+				"Proxy enable failed (%s)", strerror(-err));
+
+	return dbus_message_new_method_return(msg);
 }
 
 static DBusMessage *proxy_disable(DBusConnection *conn,
@@ -725,7 +743,6 @@ static DBusMessage *proxy_set_serial_params(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	struct serial_proxy *prx = data;
-	struct serial_adapter *adapter = prx->adapter;
 	const char *ratestr, *paritystr;
 	uint8_t databits, stopbits;
 	tcflag_t ctrl;		/* Control mode flags */
@@ -761,9 +778,6 @@ static DBusMessage *proxy_set_serial_params(DBusConnection *conn,
 	cfsetispeed(&prx->proxy_ti, speed);
 	cfsetospeed(&prx->proxy_ti, speed);
 
-	proxy_store(&adapter->src, prx->uuid128, prx->address, NULL,
-				prx->channel, 0, &prx->proxy_ti);
-
 	return dbus_message_new_method_return(msg);
 }
 
@@ -786,8 +800,8 @@ static void proxy_path_unregister(gpointer data)
 		goto done;
 
 	/* Restore the initial TTY configuration */
-	sk =  open(prx->address, O_RDWR | O_NOCTTY);
-	if (sk) {
+	sk = open(prx->address, O_RDWR | O_NOCTTY);
+	if (sk >= 0) {
 		tcsetattr(sk, TCSAFLUSH, &prx->sys_ti);
 		close(sk);
 	}
@@ -796,13 +810,13 @@ done:
 	proxy_free(prx);
 }
 
-static int register_proxy_object(struct serial_proxy *prx, char *outpath, size_t size)
+static int register_proxy_object(struct serial_proxy *prx)
 {
 	struct serial_adapter *adapter = prx->adapter;
 	char path[MAX_PATH_LENGTH + 1];
 
 	snprintf(path, MAX_PATH_LENGTH, "%s/proxy%d",
-			adapter->path, sk_counter++);
+			adapter_get_path(adapter->btd_adapter), sk_counter++);
 
 	if (!g_dbus_register_interface(adapter->conn, path,
 					SERIAL_PROXY_INTERFACE,
@@ -815,9 +829,6 @@ static int register_proxy_object(struct serial_proxy *prx, char *outpath, size_t
 	prx->path = g_strdup(path);
 	adapter->proxies = g_slist_append(adapter->proxies, prx);
 
-	if (outpath)
-		strncpy(outpath, path, size);
-
 	debug("Registered proxy: %s", path);
 
 	return 0;
@@ -825,8 +836,8 @@ static int register_proxy_object(struct serial_proxy *prx, char *outpath, size_t
 
 static int proxy_tty_register(struct serial_adapter *adapter,
 				const char *uuid128, const char *address,
-				struct termios *ti, char *outpath, size_t size,
-				gboolean save)
+				struct termios *ti,
+				struct serial_proxy **proxy)
 {
 	struct termios sys_ti;
 	struct serial_proxy *prx;
@@ -842,6 +853,7 @@ static int proxy_tty_register(struct serial_adapter *adapter,
 	prx->address = g_strdup(address);
 	prx->uuid128 = g_strdup(uuid128);
 	prx->type = TTY_PROXY;
+	adapter_get_address(adapter->btd_adapter, &prx->src);
 	prx->adapter = adapter;
 
 	/* Current TTY settings */
@@ -858,20 +870,20 @@ static int proxy_tty_register(struct serial_adapter *adapter,
 		memcpy(&prx->proxy_ti, ti, sizeof(*ti));
 	}
 
-	ret = register_proxy_object(prx, outpath, size);
-	if (ret < 0)
+	ret = register_proxy_object(prx);
+	if (ret < 0) {
 		proxy_free(prx);
+		return ret;
+	}
 
-	if (save)
-		proxy_store(&adapter->src, uuid128, address, NULL,
-			prx->channel, 0, &prx->proxy_ti);
+	*proxy = prx;
 
 	return ret;
 }
 
 static int proxy_socket_register(struct serial_adapter *adapter,
 				const char *uuid128, const char *address,
-				char *outpath, size_t size, gboolean save)
+				struct serial_proxy **proxy)
 {
 	struct serial_proxy *prx;
 	int ret;
@@ -880,22 +892,23 @@ static int proxy_socket_register(struct serial_adapter *adapter,
 	prx->address = g_strdup(address);
 	prx->uuid128 = g_strdup(uuid128);
 	prx->type = UNIX_SOCKET_PROXY;
+	adapter_get_address(adapter->btd_adapter, &prx->src);
 	prx->adapter = adapter;
 
-	ret = register_proxy_object(prx, outpath, size);
-	if (ret < 0)
+	ret = register_proxy_object(prx);
+	if (ret < 0) {
 		proxy_free(prx);
+		return ret;
+	}
 
-	if (save)
-		proxy_store(&adapter->src, uuid128, address, NULL,
-				prx->channel, 0, NULL);
+	*proxy = prx;
 
 	return ret;
 }
 
 static int proxy_tcp_register(struct serial_adapter *adapter,
 				const char *uuid128, const char *address,
-				char *outpath, size_t size, gboolean save)
+				struct serial_proxy **proxy)
 {
 	struct serial_proxy *prx;
 	int ret;
@@ -904,15 +917,16 @@ static int proxy_tcp_register(struct serial_adapter *adapter,
 	prx->address = g_strdup(address);
 	prx->uuid128 = g_strdup(uuid128);
 	prx->type = TCP_SOCKET_PROXY;
+	adapter_get_address(adapter->btd_adapter, &prx->src);
 	prx->adapter = adapter;
 
-	ret = register_proxy_object(prx, outpath, size);
-	if (ret < 0)
+	ret = register_proxy_object(prx);
+	if (ret < 0) {
 		proxy_free(prx);
+		return ret;
+	}
 
-	if (save)
-		proxy_store(&adapter->src, uuid128, address, NULL,
-				prx->channel, 0, NULL);
+	*proxy = prx;
 
 	return ret;
 }
@@ -960,16 +974,86 @@ static int proxy_pathcmp(gconstpointer proxy, gconstpointer p)
 	return strcmp(prx->path, path);
 }
 
+static int register_proxy(struct serial_adapter *adapter,
+				const char *uuid_str, const char *address,
+				struct serial_proxy **proxy)
+{
+	proxy_type_t type;
+	int err;
+
+	type = addr2type(address);
+	if (type == UNKNOWN_PROXY_TYPE)
+		return -EINVAL;
+
+	/* Only one proxy per address(TTY or unix socket) is allowed */
+	if (g_slist_find_custom(adapter->proxies, address, proxy_addrcmp))
+		return -EALREADY;
+
+	switch (type) {
+	case UNIX_SOCKET_PROXY:
+		err = proxy_socket_register(adapter, uuid_str, address, proxy);
+		break;
+	case TTY_PROXY:
+		err = proxy_tty_register(adapter, uuid_str, address, NULL,
+					proxy);
+		break;
+	case TCP_SOCKET_PROXY:
+		err = proxy_tcp_register(adapter, uuid_str, address, proxy);
+		break;
+	default:
+		err = -EINVAL;
+	}
+
+	if (err < 0)
+		return err;
+
+	g_dbus_emit_signal(adapter->conn,
+				adapter_get_path(adapter->btd_adapter),
+				SERIAL_MANAGER_INTERFACE, "ProxyCreated",
+				DBUS_TYPE_STRING, &(*proxy)->path,
+				DBUS_TYPE_INVALID);
+
+	return 0;
+}
+
+static void unregister_proxy(struct serial_proxy *proxy)
+{
+	struct serial_adapter *adapter = proxy->adapter;
+	char *path = g_strdup(proxy->path);
+
+	if (proxy->watch > 0)
+		g_dbus_remove_watch(adapter->conn, proxy->watch);
+
+	g_dbus_emit_signal(adapter->conn,
+			adapter_get_path(adapter->btd_adapter),
+			SERIAL_MANAGER_INTERFACE, "ProxyRemoved",
+			DBUS_TYPE_STRING, &path,
+			DBUS_TYPE_INVALID);
+
+	adapter->proxies = g_slist_remove(adapter->proxies, proxy);
+
+	g_dbus_unregister_interface(adapter->conn, path,
+					SERIAL_PROXY_INTERFACE);
+
+	g_free(path);
+}
+
+static void watch_proxy(DBusConnection *connection, void *user_data)
+{
+	struct serial_proxy *proxy = user_data;
+
+	proxy->watch = 0;
+	unregister_proxy(proxy);
+}
+
 static DBusMessage *create_proxy(DBusConnection *conn,
 				DBusMessage *msg, void *data)
 {
 	struct serial_adapter *adapter = data;
-	char path[MAX_PATH_LENGTH + 1];
-	const char *pattern, *address, *ppath = path;
+	struct serial_proxy *proxy;
+	const char *pattern, *address;
 	char *uuid_str;
-	proxy_type_t type;
-	uuid_t uuid;
-	int ret;
+	int err;
 
 	if (!dbus_message_get_args(msg, NULL,
 				DBUS_TYPE_STRING, &pattern,
@@ -981,49 +1065,24 @@ static DBusMessage *create_proxy(DBusConnection *conn,
 	if (!uuid_str)
 		return invalid_arguments(msg, "Invalid UUID");
 
-	bt_string2uuid(&uuid, uuid_str);
-
-	type = addr2type(address);
-	if (type == UNKNOWN_PROXY_TYPE) {
-		g_free(uuid_str);
-		return invalid_arguments(msg, "Invalid address");
-	}
-
-	/* Only one proxy per address(TTY or unix socket) is allowed */
-	if (g_slist_find_custom(adapter->proxies, address, proxy_addrcmp)) {
-		g_free(uuid_str);
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".AlreadyExist",
-						"Proxy already exists");
-	}
-
-	switch (type) {
-	case UNIX_SOCKET_PROXY:
-		ret = proxy_socket_register(adapter, uuid_str, address,
-						path, sizeof(path), TRUE);
-		break;
-	case TTY_PROXY:
-		ret = proxy_tty_register(adapter, uuid_str, address,
-				NULL, path, sizeof(path), TRUE);
-		break;
-	case TCP_SOCKET_PROXY:
-		ret = proxy_tcp_register(adapter, uuid_str, address,
-					path, sizeof(path), TRUE);
-		break;
-	default:
-		ret = -1;
-	}
-
+	err = register_proxy(adapter, uuid_str, address, &proxy);
 	g_free(uuid_str);
 
-	if (ret < 0)
-		return failed(msg, "Create object path failed");
+	if (err == -EINVAL)
+		return invalid_arguments(msg, "Invalid address");
+	else if (err == -EALREADY)
+		return g_dbus_create_error(msg, ERROR_INTERFACE ".AlreadyExist",
+						"Proxy already exists");
+	else if (err < 0)
+		return g_dbus_create_error(msg, ERROR_INTERFACE "Failed",
+				"Proxy creation failed (%s)", strerror(-err));
 
-	g_dbus_emit_signal(adapter->conn, adapter->path,
-			SERIAL_MANAGER_INTERFACE, "ProxyCreated",
-			DBUS_TYPE_STRING, &ppath,
-			DBUS_TYPE_INVALID);
+	proxy->owner = g_strdup(dbus_message_get_sender(msg));
+	proxy->watch = g_dbus_add_disconnect_watch(conn, proxy->owner,
+						watch_proxy,
+						proxy, NULL);
 
-	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &ppath,
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &proxy->path,
 					DBUS_TYPE_INVALID);
 }
 
@@ -1060,7 +1119,7 @@ static DBusMessage *remove_proxy(DBusConnection *conn,
 {
 	struct serial_adapter *adapter = data;
 	struct serial_proxy *prx;
-	const char *path;
+	const char *path, *sender;
 	GSList *l;
 
 	if (!dbus_message_get_args(msg, NULL,
@@ -1072,16 +1131,13 @@ static DBusMessage *remove_proxy(DBusConnection *conn,
 	if (!l)
 		return does_not_exist(msg, "Invalid proxy path");
 
-	g_dbus_emit_signal(conn, adapter->path,
-			SERIAL_MANAGER_INTERFACE, "ProxyRemoved",
-			DBUS_TYPE_STRING, &path,
-			DBUS_TYPE_INVALID);
-
 	prx = l->data;
-	proxy_delete(&adapter->src, prx->address);
-	adapter->proxies = g_slist_remove(adapter->proxies, prx);
 
-	g_dbus_unregister_interface(conn, path, SERIAL_PROXY_INTERFACE);
+	sender = dbus_message_get_sender(msg);
+	if (g_strcmp0(prx->owner, sender) != 0)
+		return failed(msg, "Permission denied");
+
+	unregister_proxy(prx);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -1106,7 +1162,7 @@ static void manager_path_unregister(void *data)
 
 	adapters = g_slist_remove(adapters, adapter);
 	g_slist_free(adapter->proxies);
-	g_free(adapter->path);
+	btd_adapter_unref(adapter->btd_adapter);
 	g_free(adapter);
 }
 
@@ -1123,98 +1179,104 @@ static GDBusSignalTable manager_signals[] = {
 	{ }
 };
 
-static void parse_proxy(char *key, char *value, void *data)
-{
-	struct serial_adapter *adapter = data;
-	char uuid128[MAX_LEN_UUID_STR], tmp[3];
-	char *pvalue;
-	proxy_type_t type;
-	unsigned int pos;
-	int ch, opts;
-	struct termios ti;
-	uint8_t *pti;
-
-	memset(uuid128, 0, sizeof(uuid128));
-	ch = opts = pos = 0;
-	if (sscanf(value,"%s %d 0x%04X %n", uuid128, &ch, &opts, &pos) != 3)
-		return;
-
-	/* Extracting name */
-	value += pos;
-	pvalue = strchr(value, ':');
-	if (!pvalue)
-		return;
-
-	/* FIXME: currently name is not used */
-	*pvalue = '\0';
-
-	type = addr2type(key);
-	switch (type) {
-	case TTY_PROXY:
-		/* Extracting termios */
-		pvalue++;
-		if (!pvalue || strlen(pvalue) != (2 * sizeof(ti)))
-			return;
-
-		memset(&ti, 0, sizeof(ti));
-		memset(tmp, 0, sizeof(tmp));
-
-		/* Converting to termios struct */
-		pti = (uint8_t *) &ti;
-		for (pos = 0; pos < sizeof(ti); pos++, pvalue += 2, pti++) {
-			memcpy(tmp, pvalue, 2);
-			*pti = (uint8_t) strtol(tmp, NULL, 16);
-		}
-
-		proxy_tty_register(adapter, uuid128, key, &ti, NULL, 0, FALSE);
-		break;
-	case UNIX_SOCKET_PROXY:
-		proxy_socket_register(adapter, uuid128, key, NULL, 0, FALSE);
-		break;
-	case TCP_SOCKET_PROXY:
-		proxy_tcp_register(adapter, uuid128, key, NULL, 0, FALSE);
-		break;
-	default:
-		return;
-	}
-}
-
-static void register_stored(struct serial_adapter *adapter)
-{
-	char filename[PATH_MAX + 1];
-	char address[18];
-
-	ba2str(&adapter->src, address);
-	snprintf(filename, PATH_MAX, "%s/%s/proxy", STORAGEDIR, address);
-	textfile_foreach(filename, parse_proxy, adapter);
-}
-
-static struct serial_adapter *find_adapter(GSList *list, const char *path)
+static struct serial_adapter *find_adapter(GSList *list,
+					struct btd_adapter *btd_adapter)
 {
 	GSList *l;
 
 	for (l = list; l; l = l->next) {
 		struct serial_adapter *adapter = l->data;
 
-		if (g_str_equal(adapter->path, path))
+		if (adapter->btd_adapter == btd_adapter)
 			return adapter;
 	}
 
 	return NULL;
 }
 
-int proxy_register(DBusConnection *conn, const char *path, bdaddr_t *src)
+static void serial_proxy_init(struct serial_adapter *adapter)
+{
+	GKeyFile *config;
+	GError *gerr = NULL;
+	const char *file = CONFIGDIR "/serial.conf";
+	char **group_list;
+	int i;
+
+	config = g_key_file_new();
+
+	if (!g_key_file_load_from_file(config, file, 0, &gerr)) {
+		error("Parsing %s failed: %s", file, gerr->message);
+		g_error_free(gerr);
+		g_key_file_free(config);
+		return;
+	}
+
+	group_list = g_key_file_get_groups(config, NULL);
+
+	for (i = 0; group_list[i] != NULL; i++) {
+		char *group_str = group_list[i], *uuid_str, *address;
+		int err;
+		struct serial_proxy *prx;
+
+		/* string length of "Proxy" is 5 */
+		if (strlen(group_str) < 5 || strncmp(group_str, "Proxy", 5))
+			continue;
+
+		uuid_str = g_key_file_get_string(config, group_str, "UUID",
+									&gerr);
+		if (gerr) {
+			debug("%s: %s", file, gerr->message);
+			g_error_free(gerr);
+			g_key_file_free(config);
+			return;
+		}
+
+		address = g_key_file_get_string(config, group_str, "Address",
+									&gerr);
+		if (gerr) {
+			debug("%s: %s", file, gerr->message);
+			g_error_free(gerr);
+			g_key_file_free(config);
+			g_free(uuid_str);
+			return;
+		}
+
+		err = register_proxy(adapter, uuid_str, address, &prx);
+		if (err == -EINVAL)
+			error("Invalid address.");
+		else if (err == -EALREADY)
+			debug("Proxy already exists.");
+		else if (err < 0)
+			error("Proxy creation failed (%s)", strerror(-err));
+		else {
+			err = enable_proxy(prx);
+			if (err < 0)
+				error("Proxy enable failed (%s)",
+						strerror(-err));
+		}
+
+		g_free(uuid_str);
+		g_free(address);
+	}
+
+	g_strfreev(group_list);
+	g_key_file_free(config);
+}
+
+int proxy_register(DBusConnection *conn, struct btd_adapter *btd_adapter)
 {
 	struct serial_adapter *adapter;
+	const char *path;
 
-	adapter = find_adapter(adapters, path);
+	adapter = find_adapter(adapters, btd_adapter);
 	if (adapter)
 		return -EINVAL;
 
 	adapter = g_new0(struct serial_adapter, 1);
 	adapter->conn = dbus_connection_ref(conn);
-	bacpy(&adapter->src, src);
-	adapter->path = g_strdup(path);
+	adapter->btd_adapter = btd_adapter_ref(btd_adapter);
+
+	path = adapter_get_path(btd_adapter);
 
 	if (!g_dbus_register_interface(conn, path,
 					SERIAL_MANAGER_INTERFACE,
@@ -1225,24 +1287,25 @@ int proxy_register(DBusConnection *conn, const char *path, bdaddr_t *src)
 		return -1;
 	}
 
-	register_stored(adapter);
-
 	adapters = g_slist_append(adapters, adapter);
 
 	debug("Registered interface %s on path %s",
 		SERIAL_MANAGER_INTERFACE, path);
 
+	serial_proxy_init(adapter);
+
 	return 0;
 }
 
-void proxy_unregister(const char *path)
+void proxy_unregister(struct btd_adapter *btd_adapter)
 {
 	struct serial_adapter *adapter;
 
-	adapter = find_adapter(adapters, path);
+	adapter = find_adapter(adapters, btd_adapter);
 	if (!adapter)
 		return;
 
-	g_dbus_unregister_interface(adapter->conn, path,
+	g_dbus_unregister_interface(adapter->conn,
+			adapter_get_path(btd_adapter),
 			SERIAL_MANAGER_INTERFACE);
 }
